=== modified file 'loggerhead/apps/branch.py'
--- loggerhead/apps/branch.py	2008-06-30 08:51:30 +0000
+++ loggerhead/apps/branch.py	2008-06-25 06:18:13 +0000
@@ -1,13 +1,11 @@
 import logging
 import urllib
 
-import bzrlib.branch
-import bzrlib.lru_cache
-
 from paste import request
 from paste import httpexceptions
 
 from loggerhead.apps import static_app
+
 from loggerhead.controllers.changelog_ui import ChangeLogUI
 from loggerhead.controllers.inventory_ui import InventoryUI
 from loggerhead.controllers.annotate_ui import AnnotateUI
@@ -20,30 +18,31 @@
 
 class BranchWSGIApp(object):
 
-    def __init__(self, branch_url, friendly_name=None, config={}, graph_cache=None):
+    def __init__(self, branch_url, friendly_name=None, config={}):
         self.branch_url = branch_url
+        self._history = None
         self._config = config
         self.friendly_name = friendly_name
         self.log = logging.getLogger('loggerhead.%s' % (friendly_name,))
-        if graph_cache is None:
-            graph_cache = bzrlib.lru_cache.LRUCache()
-        self.graph_cache = graph_cache
 
-    def get_history(self, b):
-        _history = History.from_branch(b, self.graph_cache)
-        cache_path = self._config.get('cachepath', None)
-        if cache_path is not None:
-            # Only import the cache if we're going to use it.
-            # This makes sqlite optional
-            try:
-                from loggerhead.changecache import FileChangeCache
-            except ImportError:
-                self.log.debug("Couldn't load python-sqlite," 
-                               " continuing without using a cache")
-            else:
-                _history.use_file_cache(
-                    FileChangeCache(_history, cache_path))
-        return _history
+    @property
+    def history(self):
+        if (self._history is None) or self._history.out_of_date():
+            self.log.debug('Reload branch history...')
+            _history = self._history = History.from_folder(self.branch_url)
+            cache_path = self._config.get('cachepath', None)
+            if cache_path is not None:
+                # Only import the cache if we're going to use it.
+                # This makes sqlite optional
+                try:
+                    from loggerhead.changecache import FileChangeCache
+                except ImportError:
+                    self.log.debug("Couldn't load python-sqlite," 
+                                   " continuing without using a cache")
+                else:
+                    _history.use_file_cache(FileChangeCache(_history, 
+                                                            cache_path))
+        return self._history
 
     def url(self, *args, **kw):
         if isinstance(args[0], list):
@@ -75,7 +74,7 @@
         }
 
     def last_updated(self):
-        h = self.get_history()
+        h = self.history
         change = h.get_changes([ h.last_revid ])[0]
         return change.date
 
@@ -97,10 +96,5 @@
         cls = self.controllers_dict.get(path)
         if cls is None:
             raise httpexceptions.HTTPNotFound()
-        b = bzrlib.branch.Branch.open(self.branch_url)
-        b.lock_read()
-        try:
-            c = cls(self, self.get_history(b))
-            return c(environ, start_response)
-        finally:
-            b.unlock()
+        c = cls(self)
+        return c(environ, start_response)

=== modified file 'loggerhead/apps/config.py'
--- loggerhead/apps/config.py	2008-06-30 09:09:16 +0000
+++ loggerhead/apps/config.py	2008-06-25 07:03:06 +0000
@@ -7,8 +7,6 @@
 import os
 import posixpath
 
-import bzrlib.lru_cache
-
 from configobj import ConfigObj
 
 from paste.request import path_info_pop
@@ -26,14 +24,13 @@
 from loggerhead.history import is_branch
 
 class Project(object):
-    def __init__(self, name, config, root_config, graph_cache):
+    def __init__(self, name, config, root_config):
         self.name = name
         self.friendly_name = config.get('name', name)
         self.description = config.get('description', '')
         self.long_description = config.get('long_description', '')
         self._config = config
         self._root_config = root_config
-        self.graph_cache = graph_cache
 
         self.views = []
         self.views_by_name = {}
@@ -80,36 +77,30 @@
             return posixpath.join(url, folder) + '/'
         return None
 
-    def _get_description(self, view, view_config, history):
+    def _get_description(self, view, view_config):
         description = view_config.get('description', None)
         if description is not None:
             return description
-        description = history._branch.get_config().get_user_option('description')
+        description = view.history._branch.get_config().get_user_option('description')
         return description
 
     def _add_view(self, view_name, view_config, folder):
-        view = BranchWSGIApp(folder, view_name, view_config, self.graph_cache)
-        b = bzrlib.branch.Branch.open(folder)
-        b.lock_read()
-        try:
-            history = view.get_history(b)
-            friendly_name = view_config.get('branch_name', None)
+        view = BranchWSGIApp(folder, view_name, view_config)
+        friendly_name = view_config.get('branch_name', None)
+        if friendly_name is None:
+            friendly_name = view.history.get_config().get_nickname()
             if friendly_name is None:
-                friendly_name = history.get_config().get_nickname()
-                if friendly_name is None:
-                    friendly_name = view_name
-            view.friendly_name = friendly_name
-            view.name = view_name
-            branch_url = self._get_branch_url(view, view_config, view_name)
-            if branch_url is not None:
-                view.branch_url = branch_url
-            view.description = self._get_description(view, view_config, history)
-            view._src_folder = folder
-            view._view_config = view_config
-            self.views.append(view)
-            self.views_by_name[view_name] = view
-        finally:
-            b.unlock()
+                friendly_name = view_name
+        view.friendly_name = friendly_name
+        view.name = view_name
+        branch_url = self._get_branch_url(view, view_config, view_name)
+        if branch_url is not None:
+            view.branch_url = branch_url
+        view.description = self._get_description(view, view_config)
+        view._src_folder = folder
+        view._view_config = view_config
+        self.views.append(view)
+        self.views_by_name[view_name] = view
 
     def call(self, environ, start_response):
         segment = path_info_pop(environ)
@@ -128,10 +119,9 @@
         self.projects = []
         self.config = config
         self.projects_by_name = {}
-        graph_cache = bzrlib.lru_cache.LRUCache()
         for project_name in self.config.sections:
             project = Project(
-                project_name, self.config[project_name], self.config, graph_cache)
+                project_name, self.config[project_name], self.config)
             self.projects.append(project)
             self.projects_by_name[project_name] = project
 

=== modified file 'loggerhead/apps/filesystem.py'
--- loggerhead/apps/filesystem.py	2008-06-30 21:15:22 +0000
+++ loggerhead/apps/filesystem.py	2008-06-18 12:27:04 +0000
@@ -2,7 +2,7 @@
 import os
 import tempfile
 
-from bzrlib import branch, errors, lru_cache
+from bzrlib import branch, errors
 
 from paste.request import path_info_pop
 from paste.wsgiwrappers import WSGIRequest, WSGIResponse
@@ -37,9 +37,9 @@
             name = os.path.basename(os.path.abspath(path))
         else:
             name = self.folder
-        branch_app = BranchWSGIApp(
-            path, name, {'cachepath': sql_dir}, self.root.graph_cache)
-        return branch_app.app
+        h = BranchWSGIApp(path, name, {'cachepath': sql_dir})
+        self.root.cache[path] = h
+        return h.app
 
     def __call__(self, environ, start_response):
         path = os.path.join(self.root.folder, self.folder)
@@ -67,7 +67,7 @@
 
 class BranchesFromFileSystemRoot(object):
     def __init__(self, folder):
-        self.graph_cache = lru_cache.LRUCache()
+        self.cache = {}
         self.folder = folder
     def __call__(self, environ, start_response):
         environ['loggerhead.static.url'] = environ['SCRIPT_NAME']

=== modified file 'loggerhead/controllers/__init__.py'
--- loggerhead/controllers/__init__.py	2008-06-30 08:38:49 +0000
+++ loggerhead/controllers/__init__.py	2008-06-25 07:03:06 +0000
@@ -55,14 +55,13 @@
 
     template_path = None
 
-    def __init__(self, branch, history):
+    def __init__(self, branch):
         self._branch = branch
-        self._history = history
         self.log = branch.log
 
     def __call__(self, environ, start_response):
         z = time.time()
-        h = self._history
+        h = self._branch.history
         kw = dict(parse_querystring(environ))
         util.set_context(kw)
 

=== modified file 'loggerhead/controllers/changelog_ui.py'
--- loggerhead/controllers/changelog_ui.py	2008-06-30 08:38:49 +0000
+++ loggerhead/controllers/changelog_ui.py	2008-06-25 00:00:19 +0000
@@ -66,7 +66,7 @@
         navigation = util.Container(
             pagesize=pagesize, revid=revid, start_revid=start_revid,
             revid_list=revid_list, filter_file_id=filter_file_id,
-            scan_url='/changes', branch=self._branch, feed=True, history=h)
+            scan_url='/changes', branch=self._branch, feed=True)
         if query is not None:
             navigation.query = query
         util.fill_in_navigation(navigation)

=== modified file 'loggerhead/controllers/revision_ui.py'
--- loggerhead/controllers/revision_ui.py	2008-06-30 08:38:49 +0000
+++ loggerhead/controllers/revision_ui.py	2008-06-25 00:00:19 +0000
@@ -52,7 +52,7 @@
         navigation = util.Container(
             revid_list=revid_list, revid=revid, start_revid=start_revid,
             filter_file_id=filter_file_id, pagesize=1,
-            scan_url='/revision', branch=self._branch, feed=True, history=h)
+            scan_url='/revision', branch=self._branch, feed=True)
         if query is not None:
             navigation.query = query
         util.fill_in_navigation(navigation)

=== removed file 'loggerhead/graphcache.py'
--- loggerhead/graphcache.py	2008-07-01 00:01:47 +0000
+++ loggerhead/graphcache.py	1970-01-01 00:00:00 +0000
@@ -1,63 +0,0 @@
-# Cache the whole history data needed by loggerhead about a branch.
-
-import logging
-import time
-
-from bzrlib.revision import is_null, NULL_REVISION
-from bzrlib.tsort import merge_sort
-
-
-def _strip_NULL_ghosts(revision_graph):
-    """
-    Copied over from bzrlib meant as a temporary workaround for
-    deprecated methods.
-    """
-    # Filter ghosts, and null:
-    if NULL_REVISION in revision_graph:
-        del revision_graph[NULL_REVISION]
-    for key, parents in revision_graph.items():
-        revision_graph[key] = tuple(parent for parent in parents if parent
-            in revision_graph)
-    return revision_graph
-
-
-def compute_whole_history_data(branch):
-    z = time.time()
-
-    last_revid = branch.last_revision()
-
-    log = logging.getLogger('loggerhead.%s' % (branch.nick,))
-
-    graph = branch.repository.get_graph()
-    parent_map = dict(((key, value) for key, value in
-         graph.iter_ancestry([last_revid]) if value is not None))
-
-    _revision_graph = _strip_NULL_ghosts(parent_map)
-    _full_history = []
-    _revision_info = {}
-    _revno_revid = {}
-    if is_null(last_revid):
-        _merge_sort = []
-    else:
-        _merge_sort = merge_sort(
-            _revision_graph, last_revid, generate_revno=True)
-
-    for (seq, revid, merge_depth, revno, end_of_merge) in _merge_sort:
-        _full_history.append(revid)
-        revno_str = '.'.join(str(n) for n in revno)
-        _revno_revid[revno_str] = revid
-        _revision_info[revid] = (
-            seq, revid, merge_depth, revno_str, end_of_merge)
-
-    _where_merged = {}
-
-    for revid in _revision_graph.keys():
-        if _revision_info[revid][2] == 0:
-            continue
-        for parent in _revision_graph[revid]:
-            _where_merged.setdefault(parent, set()).add(revid)
-
-    log.info('built revision graph cache: %r secs' % (time.time() - z,))
-
-    return (_revision_graph, _full_history, _revision_info,
-            _revno_revid, _merge_sort, _where_merged)

=== modified file 'loggerhead/history.py'
--- loggerhead/history.py	2008-07-01 00:01:47 +0000
+++ loggerhead/history.py	2008-06-23 22:22:01 +0000
@@ -36,7 +36,7 @@
 from StringIO import StringIO
 
 from loggerhead import util
-from loggerhead.graphcache import compute_whole_history_data
+from loggerhead.util import decorator
 
 import bzrlib
 import bzrlib.branch
@@ -48,6 +48,22 @@
 import bzrlib.ui
 
 
+with_branch_lock = util.with_lock('_lock', 'branch')
+
+
+@decorator
+def with_bzrlib_read_lock(unbound):
+    def bzrlib_read_locked(self, *args, **kw):
+        #self.log.debug('-> %r bzr lock', id(threading.currentThread()))
+        self._branch.repository.lock_read()
+        try:
+            return unbound(self, *args, **kw)
+        finally:
+            self._branch.repository.unlock()
+            #self.log.debug('<- %r bzr lock', id(threading.currentThread()))
+    return bzrlib_read_locked
+
+
 # bzrlib's UIFactory is not thread-safe
 uihack = threading.local()
 
@@ -175,23 +191,86 @@
 
     def __init__(self):
         self._file_change_cache = None
+        self._lock = threading.RLock()
 
     @classmethod
-    def from_branch(cls, branch, graph_cache):
+    def from_branch(cls, branch):
         z = time.time()
         self = cls()
         self._branch = branch
-        self.log = logging.getLogger('loggerhead.%s' % (branch.nick,))
-
-        self._last_revid = branch.last_revision()
-        d = graph_cache.get(self._last_revid)
-        if d is None:
-            d = compute_whole_history_data(branch)
-            graph_cache[self._last_revid] = d
-        (self._revision_graph, self._full_history, self._revision_info,
-         self._revno_revid, self._merge_sort, self._where_merged) = d
+        self._last_revid = self._branch.last_revision()
+
+        self.log = logging.getLogger('loggerhead.%s' % (self._branch.nick,))
+
+        graph = branch.repository.get_graph()
+        parent_map = dict(((key, value) for key, value in
+             graph.iter_ancestry([self._last_revid]) if value is not None))
+
+        self._revision_graph = self._strip_NULL_ghosts(parent_map)
+        self._full_history = []
+        self._revision_info = {}
+        self._revno_revid = {}
+        if bzrlib.revision.is_null(self._last_revid):
+            self._merge_sort = []
+        else:
+            self._merge_sort = bzrlib.tsort.merge_sort(
+                self._revision_graph, self._last_revid, generate_revno=True)
+
+        for (seq, revid, merge_depth, revno, end_of_merge) in self._merge_sort:
+            self._full_history.append(revid)
+            revno_str = '.'.join(str(n) for n in revno)
+            self._revno_revid[revno_str] = revid
+            self._revision_info[revid] = (
+                seq, revid, merge_depth, revno_str, end_of_merge)
+
+        # cache merge info
+        self._where_merged = {}
+
+        for revid in self._revision_graph.keys():
+            if self._revision_info[revid][2] == 0:
+                continue
+            for parent in self._revision_graph[revid]:
+                self._where_merged.setdefault(parent, set()).add(revid)
+
+        self.log.info('built revision graph cache: %r secs' % (time.time() - z,))
         return self
 
+    @staticmethod
+    def _strip_NULL_ghosts(revision_graph):
+        """
+        Copied over from bzrlib meant as a temporary workaround deprecated 
+        methods.
+        """
+
+        # Filter ghosts, and null:
+        if bzrlib.revision.NULL_REVISION in revision_graph:
+            del revision_graph[bzrlib.revision.NULL_REVISION]
+        for key, parents in revision_graph.items():
+            revision_graph[key] = tuple(parent for parent in parents if parent
+                in revision_graph)
+        return revision_graph
+
+    @classmethod
+    def from_folder(cls, path):
+        b = bzrlib.branch.Branch.open(path)
+        b.lock_read()
+        try:
+            return cls.from_branch(b)
+        finally:
+            b.unlock()
+
+    @with_branch_lock
+    def out_of_date(self):
+        # the branch may have been upgraded on disk, in which case we're stale.
+        newly_opened = bzrlib.branch.Branch.open(self._branch.base)
+        if self._branch.__class__ is not \
+               newly_opened.__class__:
+            return True
+        if self._branch.repository.__class__ is not \
+               newly_opened.repository.__class__:
+            return True
+        return self._branch.last_revision() != self._last_revid
+
     def use_file_cache(self, cache):
         self._file_change_cache = cache
 
@@ -201,6 +280,7 @@
 
     last_revid = property(lambda self: self._last_revid, None, None)
 
+    @with_branch_lock
     def get_config(self):
         return self._branch.get_config()
 
@@ -241,6 +321,7 @@
                 return
             revid = parents[0]
 
+    @with_branch_lock
     def get_short_revision_history_by_fileid(self, file_id):
         # wow.  is this really the only way we can get this list?  by
         # man-handling the weave store directly? :-0
@@ -251,6 +332,7 @@
         revids = [r for r in self._full_history if r in w_revids]
         return revids
 
+    @with_branch_lock
     def get_revision_history_since(self, revid_list, date):
         # if a user asks for revisions starting at 01-sep, they mean inclusive,
         # so start at midnight on 02-sep.
@@ -264,6 +346,7 @@
         index = -index
         return revid_list[index:]
 
+    @with_branch_lock
     def get_search_revid_list(self, query, revid_list):
         """
         given a "quick-search" query, try a few obvious possible meanings:
@@ -323,6 +406,7 @@
             revid = self._revno_revid[revid]
         return revid
 
+    @with_branch_lock
     def get_file_view(self, revid, file_id):
         """
         Given a revid and optional path, return a (revlist, revid) for
@@ -342,6 +426,7 @@
             revlist = list(self.get_revids_from(None, revid))
         return revlist
 
+    @with_branch_lock
     def get_view(self, revid, start_revid, file_id, query=None):
         """
         use the URL parameters (revid, start_revid, file_id, and query) to
@@ -394,9 +479,11 @@
         else:
             return None, None, []
 
+    @with_branch_lock
     def get_inventory(self, revid):
         return self._branch.repository.get_revision_inventory(revid)
 
+    @with_branch_lock
     def get_path(self, revid, file_id):
         if (file_id is None) or (file_id == ''):
             return ''
@@ -405,6 +492,7 @@
             path = '/' + path
         return path
 
+    @with_branch_lock
     def get_file_id(self, revid, path):
         if (len(path) > 0) and not path.startswith('/'):
             path = '/' + path
@@ -483,6 +571,7 @@
                 else:
                     p.branch_nick = '(missing)'
 
+    @with_branch_lock
     def get_changes(self, revid_list):
         """Return a list of changes objects for the given revids.
 
@@ -509,6 +598,8 @@
 
         return changes
 
+    @with_branch_lock
+    @with_bzrlib_read_lock
     def get_changes_uncached(self, revid_list):
         # FIXME: deprecated method in getting a null revision
         revid_list = filter(lambda revid: not bzrlib.revision.is_null(revid),
@@ -579,6 +670,7 @@
 
         return [self.parse_delta(delta) for delta in delta_list]
 
+    @with_branch_lock
     def get_file_changes(self, entries):
         if self._file_change_cache is None:
             return self.get_file_changes_uncached(entries)
@@ -591,6 +683,7 @@
         for entry, changes in zip(entries, changes_list):
             entry.changes = changes
 
+    @with_branch_lock
     def get_change_with_diff(self, revid, compare_revid=None):
         change = self.get_changes([revid])[0]
 
@@ -609,6 +702,7 @@
 
         return change
 
+    @with_branch_lock
     def get_file(self, file_id, revid):
         "returns (path, filename, data)"
         inv = self.get_inventory(revid)
@@ -741,6 +835,7 @@
             for m in change.changes.modified:
                 m.sbs_chunks = _make_side_by_side(m.chunks)
 
+    @with_branch_lock
     def get_filelist(self, inv, file_id, sort_type=None):
         """
         return the list of all files (and their attributes) within a given
@@ -793,6 +888,7 @@
 
     _BADCHARS_RE = re.compile(ur'[\x00-\x08\x0b\x0e-\x1f]')
 
+    @with_branch_lock
     def annotate_file(self, file_id, revid):
         z = time.time()
         lineno = 1

=== modified file 'loggerhead/templates/revisionfilechanges.pt'
--- loggerhead/templates/revisionfilechanges.pt	2008-06-27 04:32:15 +0000
+++ loggerhead/templates/revisionfilechanges.pt	2007-09-05 18:40:01 +0000
@@ -24,7 +24,11 @@
     <td class="files"
         tal:define="filename python:removed[0];
                     file_id  python:removed[1]">
-      <tal:block content="filename" />
+      <a tal:attributes="href  python:url(['/annotate', entry.revid], file_id=file_id);
+                         title string:Annotate ${filename}"
+         tal:content="filename" class="filename">
+        filename
+      </a>
     </td>
   </tr>
   <tr tal:repeat="renamed entry/changes/renamed">

=== modified file 'loggerhead/tests/test_simple.py'
--- loggerhead/tests/test_simple.py	2008-07-01 00:01:47 +0000
+++ loggerhead/tests/test_simple.py	2008-06-18 12:27:04 +0000
@@ -51,6 +51,9 @@
             folder = '%(branch)s'
     """
 
+    def makeHistory(self):
+        return History.from_folder(self.bzrbranch)
+
     def setUpLoggerhead(self):
         app = TestApp(BranchWSGIApp(self.bzrbranch).app)
         return app

=== modified file 'loggerhead/util.py'
--- loggerhead/util.py	2008-06-30 08:38:49 +0000
+++ loggerhead/util.py	2008-06-25 06:03:26 +0000
@@ -356,11 +356,11 @@
 
     navigation.prev_page_revid = get_offset(-1 * navigation.pagesize)
     navigation.next_page_revid = get_offset(1 * navigation.pagesize)
-    prev_page_revno = navigation.history.get_revno(
+    prev_page_revno = navigation.branch.history.get_revno(
             navigation.prev_page_revid)
-    next_page_revno = navigation.history.get_revno(
+    next_page_revno = navigation.branch.history.get_revno(
             navigation.next_page_revid)
-    start_revno = navigation.history.get_revno(navigation.start_revid)
+    start_revno = navigation.branch._history.get_revno(navigation.start_revid)
 
     params = { 'filter_file_id': navigation.filter_file_id }
     if getattr(navigation, 'query', None) is not None:

=== modified file 'start-loggerhead.py'
--- start-loggerhead.py	2008-06-30 23:24:32 +0000
+++ start-loggerhead.py	2008-06-23 01:40:28 +0000
@@ -50,9 +50,9 @@
     access_log.setLevel(logging.INFO)
     access_log.setFormatter(f)
 
-    logging.getLogger('').setLevel(logging.DEBUG)
     logging.getLogger('').addHandler(debug_log)
-    logging.getLogger('wsgi').addHandler(access_log)
+    logging.getLogger('turbogears.access').addHandler(access_log)
+    logging.getLogger('turbogears.controllers').setLevel(logging.INFO)
 
     if foreground:
         logging.getLogger('').addHandler(stdout_log)

